import os from 'node:os';
import { execFile } from 'node:child_process';
import path from 'node:path';
import fs, { createReadStream } from 'node:fs';
import { createHash } from 'node:crypto';

function mapArchAndPlatform(platform, arch, ver = "") {
	if (platform === "linux") {
		switch (arch) {
			case "x64":
				return {file: "linux-x64", link: `node-${ver}-linux-x64`}

			case "arm64":
				return {file: "linux-arm64", link: `node-${ver}-linux-arm64`}
		}
	} else if (platform === "darwin") {
		switch (arch) {
			case "x64":
				return {file: "osx-x64-tar", link: `node-${ver}-darwin-x64`}

			case "arm64":
				return {file: "osx-arm64-tar", link: `node-${ver}-darwin-arm64`}
		}
	}

	throw new Error(`Unsupported platform/arch combination.`)
}

async function impl_getAvailableVersions() {
	let ret = [];

	const arch_platform_identifier = mapArchAndPlatform(os.platform(), os.arch()).file;

	const response = await fetch("https://nodejs.org/dist/index.tab");
	const list = await response.text();
	// first line is description
	const entries = list.split("\n").slice(1);

	for (const entry of entries) {
		const fields = entry.split("\t");

		if (2 >= fields.length) continue

		const [version, date, files_str] = fields;
		const files = files_str.split(",");

		if (!files.includes(arch_platform_identifier)) continue

		ret.push(version);
	}

	return ret
}

async function impl_getDownloadURL(version) {
	const arch_platform_identifier = mapArchAndPlatform(
		os.platform(), os.arch(), version
	).link;

	return `https://nodejs.org/dist/${version}/${arch_platform_identifier}.tar.gz`
}

function createRandomIdentifier(len) {
	let str = "";
	let bytes = [];

	if (typeof crypto === "object" && typeof crypto.getRandomValues === "function") {
		bytes = new Uint8Array(len);

		crypto.getRandomValues(bytes);
	} else {
		for (let i = 0; i < len; ++i) {
			// will never exceed 255
			bytes.push(Math.floor(Math.random() * 256));
		}
	}

	for (let i = 0; i < len; ++i) {
		str += (bytes[i]).toString(16).slice(0, 1);
	}

	// sanity check output
	if (str.length !== len) {
		throw new Error(`createRandomIdentifier(${len}) failed for unknown reasons.`)
	}

	return str
}

function calculateChecksum(file_path) {
	return new Promise((resolve, reject) => {
		const hash = createHash("sha256");

		hash.setEncoding("hex");

		createReadStream(file_path)
		.pipe(hash)
		.on("finish", () => {
			resolve(hash.read());
		})
		.on("error", reject);
	})
}

//
// todo: replace with fetch()
//
function downloadFileViaCURL(url) {
	return new Promise((resolve, reject) => {
		const file_dest_path = path.join(
			os.tmpdir(), createRandomIdentifier(32)
		);

		execFile("curl", [
			url, "--output", file_dest_path
		], {
			stdio: "pipe"
		}, (error, stdout, stderr) => {
			if (error) {
				reject(error);
			} else {
				resolve(file_dest_path);
			}
		});
	})
}

function extractArchiveWithTar(tar_file) {
	return new Promise((resolve, reject) => {
		const dir_dest_path = path.join(
			os.tmpdir(), createRandomIdentifier(32)
		);

		fs.mkdirSync(dir_dest_path);

		execFile("tar", [
			"-xzf", tar_file, "-C", "."
		], {
			cwd: dir_dest_path,
			stdio: "pipe"
		}, (error, stdout, stderr) => {
			if (error) {
				reject(error);
			} else {
				resolve(dir_dest_path);
			}
		});
	})
}

async function getChecksums(version, arch_platform_identifier) {
	//
	const response = await fetch(`https://nodejs.org/dist/${version}/SHASUMS256.txt`);
	const body = await response.text();
	const entries = body.split("\n");

	let target_checksum = "";

	for (const entry of entries) {
		const fields = entry.trim().split(" ").filter(p => p.length);

		if (fields.length !== 2) continue

		const [checksum, file_name] = fields;

		if (file_name === `${arch_platform_identifier}.tar.gz`) {
			target_checksum = checksum;

			break
		}
	}

	if (target_checksum.length !== 64) {
		throw new Error(`Unable to get correct checksum for download.`)
	}

	return target_checksum
}

//
// todo: cleanup on error
//
async function impl_download(version) {
	const versions = await impl_getAvailableVersions();

	if (!versions.includes(version)) {
		throw new Error(`No such version '${version}'.`)
	}

	const arch_platform_identifier = mapArchAndPlatform(
		os.platform(), os.arch(), version
	).link;

	const file_checksum = await getChecksums(version, arch_platform_identifier);

	const url = await impl_getDownloadURL(version);
	const tmp_file = await downloadFileViaCURL(url);

	const local_checksum = await calculateChecksum(tmp_file);

	if (file_checksum !== local_checksum) {
		throw new Error(`Checksum error <${file_checksum} != ${local_checksum}>.`)
	}

	const tmp_dir = await extractArchiveWithTar(tmp_file);

	fs.unlinkSync(tmp_file);

	return path.join(
		tmp_dir, arch_platform_identifier
	)
}

const getAvailableVersions = impl_getAvailableVersions;
const getDownloadURL = impl_getDownloadURL;
const download = impl_download;

var index = {
	getAvailableVersions,
	getDownloadURL,
	download
};

export { index as default, download, getAvailableVersions, getDownloadURL };
